#include "jtypes.h"
#include "jproxy.h"
#include "jhook.h"
#include "utils.h"
#include "line_buffer.h"
#include "jlog.h"


#define HTTP_TERM   "/x0D/x0A"


#define CHUNK_NO_DATA (-1)
#define CHUNK_ERROR   (-2)
#define CHUNK_ZERO    ( 0)
#define CHUNK_OK      ( 1) // used in data only request

/* get chunk size from line buffer */
static int get_chunk_size(proxy_session_t *session, int peer)
{
  char *data, *tmp;
  size_t len, rest, chunk_len;

  assert( peer == PEER_CLIENT || peer == PEER_SERVER );

  /* get buffer data */
  data = get_header_line(session, &len, peer);
  if( !data )
  {
    int ret = put_data(session, data, len, peer);
    assert( !ret );
    return CHUNK_NO_DATA;
  }

  tmp = skip_lws(data, len, &rest);
  assert( rest > 0 ); // on empty line..TODO: not right assert..

  chunk_len = line_strtol(data); 
  if( chunk_len < 0 )
    return CHUNK_ERROR;

  return chunk_len;
}







/* Get chunk data of expected size..
 * return int new_df .. expected data with CR LF (expected + 2) 
 * return value: CHUNK_ERROR, CHUNK_NO_DATA, CHUNK_OK 
 */
static int get_chunk_data(proxy_session_t *session, size_t expected, df_t **new_df, int peer)
{
  size_t len, refill_len, real_expected=expected + 2;
  char *data, *refill;

  assert(peer == PEER_CLIENT || peer == PEER_SERVER );
  
  *new_df = NULL;
  
  data = get_body_data(session, &len, peer);
  if( !data )
    return CHUNK_NO_DATA; 

  
  if( len < real_expected )
  {
    int ret = put_data(session, data, len, peer);
    assert( !ret );
    return CHUNK_NO_DATA;
  }

  /* refill buffer with overflow data */
  refill = data + real_expected;    
  refill_len = len - real_expected;
  
  int ret = put_data(session, refill, refill_len, peer); 
  assert( !ret );
 
  memset(refill, 0x00, refill_len); // gdb debug helper
  
  if(*(refill - 2) != CR || *(refill - 1) != LF)
  {
    jfree(data);
    return CHUNK_ERROR;
  }

  df_t *df = alloc_df(data, real_expected, DF_CHUNK, 0);
  if( !df )
  {
    jfree(data);
    return CHUNK_ERROR;
  }
  
  *new_df = df;
  return CHUNK_OK;
}









/* return: CHUNK_ERROR|CHUNK_NO_DATA|CHUNK_ZERO && new_df == 0    OR    CHUNK_OK && new_df with chunked data */
static int get_chunk(proxy_session_t *session, df_t **new_df, int peer)  // main get chunk function
{
  int ret,expected;
  struct client_request *req = get_cur_request(session, peer);
  
  assert( req->__status & __MSG_CUR_BODY && req->__body_type & __MSG_BODY_HAS_CHUNK);
  
  if( req->__body_status & __MSG_BODY_CHUNK_SIZE_EXPECTED )
  { 
    ret =  get_chunk_size(session, peer);
    if( ret == CHUNK_ERROR || ret == CHUNK_NO_DATA )
      return ret;

    assert( ret >= 0 );

    set_request_body_state(req, __MSG_BODY_CHUNK_CUR_PROC);
    req->chunk_len = ret;

    if( ret == 0 ) // end chunk 
    {
      set_request_body_state(req, __MSG_BODY_CHUNK_END);
      return CHUNK_ZERO;
    }
  }

  expected = req->chunk_len; 
  
  if( req->__body_status & __MSG_BODY_CHUNK_CUR_PROC )
  {
    ret = get_chunk_data(session, expected, new_df, peer); 
    if( ret == CHUNK_ERROR || ret == CHUNK_NO_DATA )
      return ret;

    set_request_body_state(req, __MSG_BODY_CHUNK_SIZE_EXPECTED);
    req->chunk_len = 0;

    return ret;

  }

  assert(0);
  return CHUNK_ERROR; // shutup gcc:)

}



int get_raw_body(proxy_session_t *session, df_t **new_df, int peer)
{
  char *data;
  size_t len, max_expected;
  df_t *df=NULL;
  
  struct client_request *req = get_cur_request(session, peer);
  assert( req->__status & __MSG_CUR_BODY && req->__body_type & __MSG_BODY_HAS_CONTENT_LENGTH);

  *new_df = NULL;

  max_expected = req->total_len;

  data = get_body_data(session, &len, peer); 
  if( !data )
    return PROTOCOL_NO_DATA;

  if( len > max_expected )
  {
    char *refill;
    size_t refill_len;

    refill = data + max_expected;
    refill_len = len - max_expected;
    
    int ret = put_data(session, refill, refill_len, peer);
    assert( !ret );

    memset(refill, 0x00, refill_len); // gdb debug ..

    len = max_expected;
  }

  req->total_len -= len;

  df =  alloc_df(data, len, DF_RAW, 0);
  if( !df )
    return PROTOCOL_ERROR;
  
  *new_df =  df;


  if( req->total_len == 0 )
    return PROTOCOL_END;
  else
    return PROTOCOL_OK;
}


int get_raw_body_on_close(proxy_session_t *session, df_t **new_df, int peer)
{
  char *data;
  size_t len;
  df_t *df=NULL;


  struct client_request *req = get_cur_request(session, peer);
  assert( (req->__status & __MSG_CUR_BODY) && 
          (req->__body_type & (__MSG_BODY_ON_CLOSE|__MSG_BODY_HAS_CONTENT_LENGTH|__MSG_BODY_HAS_CHUNK)) == __MSG_BODY_ON_CLOSE );

  assert(req->total_len == 0xFFFFFFFFUL);
  
  data = get_body_data(session, &len, peer);
  if( !data )
    return PROTOCOL_NO_DATA;

 
  df = alloc_df(data, len, DF_RAW, DF_F_ON_CLOSE);
  if( !df )
    return PROTOCOL_ERROR;

  *new_df = df;

  return PROTOCOL_OK;
}





/* modify request after body token processing */
int request_body_modifier(proxy_session_t *session, df_t *df, int peer)
{
  struct client_request *req = get_cur_request(session, peer);
  
  if( req->__body_type & __MSG_BODY_HAS_CHUNK )  
  {
    if( is_empty_line(df->data, df->size) )
    {
      assert( df->flags & DF_F_EOF );
      set_request_status(req, __MSG_COMPLETED );
    
      return 1;
    }
  }
  else if( req->__body_type & __MSG_BODY_HAS_CONTENT_LENGTH )
  {
    assert( !df->flags );
    if( req->total_len == 0 )
    {
      set_request_status(req, __MSG_COMPLETED );
      return 1;
    }
  }
  else if( req->__body_type & __MSG_BODY_ON_CLOSE )
  {
    // null operation.. to use assert above..
    return 0;
  }
  else
  {
    assert(0);
  }
    
  return 0;
}






int process_body(proxy_session_t *session, df_t **new_df, int peer)
{
  int ret;
  df_t *df=NULL;
  struct client_request *req = get_cur_request(session, peer);

  if( !req )
    return 0;
  
  *new_df = NULL;
  
  assert( req->__status & __MSG_CUR_BODY );

  if( req->__body_type & __MSG_BODY_HAS_CONTENT_LENGTH )
  {
    return  get_raw_body(session, new_df, peer);
  }
  else if ( req->__body_type & __MSG_BODY_HAS_CHUNK )
  {

    if( req->__body_status & __MSG_BODY_CHUNK_END )
    {
      size_t len;
      char *data;
      
      data = get_header_line(session, &len, peer);
      if(!data)
        return PROTOCOL_NO_DATA;      

      if( is_empty_line(data, len) ) 
      {
        df = alloc_df(data, len, DF_CHUNK, DF_F_EOF);
        if( !df )
          return PROTOCOL_ERROR;

        *new_df = df;
        return PROTOCOL_END;
      }
      else
        return PROTOCOL_ERROR;
      
    }      
    else
    {
      ret = get_chunk(session, new_df, peer);
      if( ret == CHUNK_ERROR )
        return PROTOCOL_ERROR;

      if( ret == CHUNK_NO_DATA )
        return PROTOCOL_NO_DATA;

      if( ret == CHUNK_ZERO || ret == CHUNK_OK )
        return PROTOCOL_AGAIN;
    }
  }
  else if ( req->__body_type & __MSG_BODY_ON_CLOSE )
  {
    return get_raw_body_on_close(session, new_df, peer);
  }
  else 
  {
    assert(0);
  }

  assert(0);
  return PROTOCOL_ERROR;
  
}


