#include <stdio.h>
#include <assert.h>
#include <string.h>
#include "jsocket.h"
#include "jselect.h"
#include "configure.h"
#include "jproxy.h"
#include "jhook.h"
#include "utils.h"
#include "jdns.h"
#include "jlog.h"
#include "df.h"


#define MAX_HOOK 16


/* per child global... no race */
proxy_session_t g_session;
ip_entry_t *g_ip_entry=NULL;

static int write_whole_out(j_socket *jsock, const char *buffer, size_t len)
{
  int ret;
  size_t write_data = 0;
  do 
  {
    assert(write_data <= len); // debug
    
    ret = js_write(jsock, buffer, len);
    if( ret < 0  ) // error
      return 0;
    write_data += (size_t)ret;
    len -= ret;

  } while(write_data < len);

  return 1;
}

static int __forward_to(df_t* df, j_socket *jsock )
{
  if( df->ancillary )
    __forward_to(df->ancillary, jsock); // for ancillary data recurisve call __forward_to()
        
  if( df->type == DF_CHUNK ) // forward chunk size..
  {
    size_t size;
    char *line = line_get_chunk_size(df->size-2, &size);
    char *str = line_get_str(line);

    D_LOG(MSG_DEBUG, "Appending Chunked Size: %s", str);
    
    jfree(str);

    
    write_whole_out(jsock, line, size);

    jfree(line);
    
  }

  if( df->type == DF_HEADER )
  {
    char *str = line_get_str(df->data);
    D_LOG(MSG_DEBUG, "Forwarding : HEADER: data %s", str);
    jfree(str);
    
  }
  else if( df->type == DF_RAW )
  {
    D_LOG(MSG_DEBUG, "Forwarding : RAW DATA, len: %d", df->size);
  }
  else if( df->type == DF_CHUNK )
  {
    D_LOG(MSG_DEBUG, "Forwarding : CHUNK DATA, len: %d", df->size);
  }
  
  return write_whole_out(jsock, df->data, df->size); // write all real data
}


static int forward_to_server(proxy_session_t *session, df_t* df)
{
  assert( session && df && df->data );
          
  j_socket *server = session->server_link;
  int ret = __forward_to(df, server);

  free_df(df);
  return ret;
}

static int forward_to_client(proxy_session_t *session, df_t* df)
{
  assert( session && df && df->data );
  
  j_socket *client = session->client_link;
  int ret = __forward_to(df, client);

  free_df(df);
  return ret;
}

/* client request chain function */
void init_request_chain(struct client_request_chain *chain)
{
  chain->head = chain->tail = NULL;
  chain->request_num = 0;
}

void init_request(struct client_request *request)
{
  memset(request, 0x00, sizeof(struct client_request));
  request->__status |= __MSG_CUR_HEADER | __MSG_CUR_CLIENT;
}


void  put_request_chain(struct client_request_chain *chain, struct client_request *cli_req)
{
  assert( (chain->head == NULL && chain->tail == NULL) || (chain->head != NULL && chain->tail != NULL) );
  assert( cli_req->next == NULL && cli_req->prev == NULL ); 

  D_LOG(MSG_DEBUG, "Add New Element To client_request_chain");
 
  if( chain->head == NULL )  // no entry
  {
    chain->head = cli_req;
    chain->tail = cli_req;
    
    cli_req->prev = NULL;
    cli_req->next = NULL;
  }
  else
  {
    cli_req->prev = NULL;
    cli_req->next = chain->head;
    
    chain->head = cli_req;
    cli_req->next->prev = cli_req;
  }

  chain->request_num++;
}

struct client_request *get_request_chain(struct client_request_chain *chain)
{
  assert( (chain->head == NULL && chain->tail == NULL) || (chain->head != NULL && chain->tail != NULL) );
  struct client_request *tmp; 

 
  if( chain->head == NULL ) // empty
  {
    return NULL;
  }
  else if( chain->head == chain->tail ) // one entry 
  {
    tmp = chain->head;
    chain->head = chain->tail = NULL;
    tmp->next = tmp->prev = NULL;
  }
  else // more entry
  {
    tmp = chain->tail;
    tmp->prev->next = NULL;
    chain->tail = tmp->prev;
    
    /* debug */
    assert( tmp->prev != NULL && chain->tail != NULL);
    tmp->next = tmp->prev = NULL;
  }

  chain->request_num--;

  D_LOG(MSG_DEBUG, "Getting a Request From Chain..");
  return tmp;
}

#define LINE_BUFFER_LEN 0x40000 // 2^12 
#define HTTP_LINE_TERM  "\x0D\x0A"
static int init_session(proxy_session_t *session)
{
  memset(session, 0x00, sizeof(proxy_session_t));

  session->cli_req_chain = jmalloc(sizeof(struct client_request_chain));
  if( !session->cli_req_chain )
    return 0;
  
  init_request_chain(session->cli_req_chain);

  session->lb_client = create_lb(LINE_BUFFER_LEN, HTTP_LINE_TERM);
  if( !session->lb_client )
  {
    jfree(session->cli_req_chain);
    return 0;
  }

  session->lb_server = create_lb(LINE_BUFFER_LEN, HTTP_LINE_TERM);
  if( !session->lb_server )
  {
    jfree(session->lb_client);
    jfree(session->cli_req_chain);
    return 0;
  }
  
  
  // use global chain
  session->client_chain = &g_client_filter_chain;
  session->server_chain = &g_server_filter_chain;

  return 1;
}


/* TODO: memleak on client_request if server shutdown before answering all messages */
static void fini_session(proxy_session_t *session)
{
  assert( session && session->cli_req_chain && session->lb_server && session->lb_client );
  
  jfree(session->cli_req_chain);
  destroy_lb(session->lb_client);
  destroy_lb(session->lb_server);
  
  if( session->client_link )
  {
    js_close(session->client_link);
    js_destroy_sock_tcp(session->client_link);
  }

  if( session->server_link )
  {
    js_close(session->server_link);
    js_destroy_sock_tcp(session->server_link);
  }

  memset(session, 0x00, sizeof(proxy_session_t));
}


#define OK_RESOLV      (0)  // ok
#define NO_MORE_DATA   (-1) // retry
#define NO_PARSE       (-2) // error
#define NO_RESOLV      (-3) // error

int parse_buffer_dns(const char *host, struct sockaddr_in *peer, size_t len)
{
  char *end;
  int i;
  size_t rest;

  // TODO: use URL rfc max lenght .. 
  char buffer[1024];
  assert( *host == 'H' );
  memset(buffer, 0x00, sizeof(buffer));
  
  end = line_strmem(host, HTTP_LINE_TERM, len, &rest);
  if( end == NULL)
    return NO_MORE_DATA;

  /* get resolve and save in peer args */
  end = line_strmem(host, " ", len, &rest);
  if( end == NULL )
  {
    D_LOG(MSG_ERROR, "Unable to Parse Host"); 
    return NO_PARSE;
  }
  /* skeep space */ 
  while( *end == 0x20 || *end == 0x09 )
    end++;
  
  for(i=0; *end != 0x0d && *end != 0x20 && *end != ':' && *end != 0x09 && i<sizeof(buffer)-1; i++, end++)
    buffer[i] = *end;

  if( !dns_get_ip(buffer, peer) )
  {
    D_LOG(MSG_ERROR, "Unable to Resolve Host"); 
    return NO_RESOLV;
  }
  return OK_RESOLV;
}



static int try_to_resolv_host(struct sockaddr_in *peer,  // peer to fill
                              j_socket *jsock,           // reading sock
                              proxy_session_t *session)  // current session
{
  char *tmp, *host; 
  char buffer[1024];
  int ret;
  size_t len, rest;

  memset(peer, 0x00, sizeof(struct sockaddr_in));
  
  do
  {
    ret = js_read(jsock, buffer, sizeof(buffer));
    if(ret <= 0) // close or error
    {
      D_LOG(MSG_ERROR, "Error while trying to get Host message header");
      return 0;
    }

    lb_insert_data(session->lb_client, buffer, ret, &len);    
    if( len ) // buffer full
    {
      D_LOG(MSG_ERROR, "Error while trying to get Host message header");
      return 0;
    }

    tmp = lb_query_last(session->lb_client, &len);
    
    host = line_strmem(tmp, "Host:", len, &rest);
    if( host == NULL ) 
    {
      lb_insert_data(session->lb_client, tmp, len, &len); // reinsert data
      assert( !len );
      jfree(tmp);
      continue;
    }

    ret = parse_buffer_dns(host, peer, rest); 
    if( ret == NO_MORE_DATA )
    {
      lb_insert_data(session->lb_client, tmp, len, &len); // reinsert data
      assert( !len );
      jfree(tmp);
      continue;
    }
    else if( ret == NO_RESOLV || ret == NO_PARSE )
    {
      jfree(tmp);
      return 0;
    }
    else // OK_RESOLV
      break;

  } while (1);

  lb_insert_data(session->lb_client, tmp, len, &len); // reinsert data
  assert( !len );
  jfree(tmp);

  return 1;
}


/* this function call underlaing IO multiplexig and return a set of active socket */
#define IS_IO_TIMEOUT(ret) ( ((int)(ret)) == -1 )
static j_socket_set_t *multiplex_session(proxy_session_t *session)
{
  int ret;
  static j_socket_set_t *set=NULL;
  j_socket_set_t *active;

  if(set == NULL) // first call..
  {
    set = get_j_socket_set(16);
    if( !set )
      return NULL;
    
    if( !add_j_socket_to_set(set, session->client_link) || !add_j_socket_to_set(set, session->server_link) )
      return NULL;
  }

  ret = js_select(set, &active, 60);
  switch( ret )
  {
    case 0:
      return (j_socket_set_t*)(-1);
    case -1:
      return NULL;
    
    default:
      return active;
  }
}



static int parse_server_message(proxy_session_t *session)
{
  df_t *df; 
  int ret, hook_ret;
  unsigned int status,body,completed=0;

  do 
  { 
    /* tmp value needed by prev cicle */
    struct client_request *req = get_cur_request(session, PEER_SERVER);
    if( !req )
      return 0;

    status = req->__status;
    body   = req->__body_type;
    
    if( status & __MSG_CUR_HEADER )
    {
      ret = process_header(session, &df, PEER_SERVER); 
      if( ret == PROTOCOL_ERROR )
        return 0;

      if( df ) 
      {
        request_header_modifier(session, df, PEER_SERVER);
        
        hook_ret = call_hook_chain(&g_server_filter_chain, &df, req, g_ip_entry);
        if( hook_ret < 0)
        {
          D_LOG(MSG_DEBUG, "Hook Fatal Error: desotying current session");
          return PROTOCOL_ERROR;
        }
      }
    }
    else if( status & __MSG_CUR_BODY )
    {
      ret = process_body(session, &df, PEER_SERVER);
      if( ret == PROTOCOL_ERROR )
        return 0;

      if( df )
      {
        request_body_modifier(session, df, PEER_SERVER);

        hook_ret = call_hook_chain(&g_server_filter_chain, &df, req, g_ip_entry);
        if( hook_ret < 0)
        {
          D_LOG(MSG_DEBUG, "Hook Fatal Error: desotying current session");
          return PROTOCOL_ERROR;
        }
    
      }
    }
    else if( status & __MSG_COMPLETED )
    {
      df = NULL;
      session->cur_req_to_process = NULL;

      D_LOG(MSG_DEBUG, "Freeing Client_Request: %p", req);
      jfree(req);
      
      ret = PROTOCOL_END;
      completed = 1;
    }
    else
    {
      assert(0);
    }

    /* forward data fragment */
    if( df )
      forward_to_client(session, df);
    
  }
  while( ret == PROTOCOL_AGAIN ||  // header/body return again
         (ret == PROTOCOL_END &&  !completed)  
       ); 

  /* here if ret = PROTOCOL_NO_DATA or PROTOCOL_OK */
  
  return 1;

}



static int parse_client_message(proxy_session_t *session)
{
  df_t *df; 
  int ret, hook_ret;
  unsigned int status,body,completed=0;

  do 
  { 
    /* tmp value needed by prev cicle */
    struct client_request *req = get_cur_request(session, PEER_CLIENT);
    if( !req )
      return 0;

    status = req->__status;
    body   = req->__body_type;
    
    if( status & __MSG_CUR_HEADER )
    {
      ret = process_header(session, &df, PEER_CLIENT); 
      if( ret == PROTOCOL_ERROR )
        return 0;
      
      if( df )
      {
        request_header_modifier(session, df, PEER_CLIENT);

        hook_ret = call_hook_chain(&g_client_filter_chain, &df, req, g_ip_entry);
        if( hook_ret < 0)
        {
          D_LOG(MSG_DEBUG, "Hook Fatal Error: desotying current session");
          return PROTOCOL_ERROR;
        }
      } // if(df)
    }
    else if( status & __MSG_CUR_BODY )
    {
      ret = process_body(session, &df, PEER_CLIENT);
      if( ret == PROTOCOL_ERROR )
        return 0;

      if( df )
      {
        request_body_modifier(session, df, PEER_CLIENT);

        hook_ret = call_hook_chain(&g_client_filter_chain, &df, req, g_ip_entry);
        if( hook_ret < 0)
        {
          D_LOG(MSG_DEBUG, "Hook Fatal Error: desotying current session");
          return PROTOCOL_ERROR;
        }
      } // if(df)
    }
    else if( status & __MSG_COMPLETED )
    {
      df = NULL;
      session->cur_req = NULL;
      req->__body_type = req->flags = req->__body_status = 0;
      req->__status = __MSG_CUR_HEADER | __MSG_CUR_SERVER;
      put_request_chain(session->cli_req_chain, req); 

      ret = PROTOCOL_END; // useless?
      completed = 1;
    }
    else
    {
      assert(0);
    }

    /* forward data fragment */
    if( df ) // no data or hook remove the data?
      forward_to_server(session, df);
    
  }
  while( ret == PROTOCOL_AGAIN ||  // header/body return again
         (ret == PROTOCOL_END &&  !completed)  
       ); 

  return 1;

}




static int dispatch_client_data(char *data, size_t len, proxy_session_t *session)
{
  size_t discard;
  //int ret;
  //D_LOG(MSG_DEBUG, "OK");
  discard = put_data(session, data, len, PEER_CLIENT);
  if( discard )
  {
    D_LOG(MSG_ERROR, "Buffer Full: loosing %d bytes of data: malformed request?", discard);
  }
  
  // here process the message..
  return parse_client_message(session);
}



static int dispatch_server_data(char *data, size_t len, proxy_session_t *session)
{
  size_t discard;
  //int ret;
  //D_LOG(MSG_DEBUG, "OK");
  discard = put_data(session, data, len, PEER_SERVER);
  if( discard )
  {
    D_LOG(MSG_ERROR, "Buffer Full: loosing %d bytes of data: malformed request?", discard);
  }

  // here process the message..
  //ret = parse_server_message(session);
  
  //df_t *df = alloc_df(data, len, 0, 0);
  //forward_to_client(session, df);
  
  return parse_server_message(session);
}

/* return the last active handler result or the first error/close that occurs..
 * if a close occur try to complete active set of request anyway.. 
 */
#define DISPATCH_SESSION_OK     (1)
#define DISPATCH_SESSION_CLOSE  (0)
#define DISPATCH_SESSION_ERROR  (-1)

#define MESSAGE_CHUNK_SIZE 1024
static int dispatch_session_event(j_socket_set_t *active, proxy_session_t *session)
{
  j_socket *jsock_active;
  j_socket *client = session->client_link, *server = session->server_link;
  
  rewind_j_socket_set(active);

  while( (jsock_active = get_next_j_socket(active)) != NULL)
  {
    int ret;
    
    char *data = jmalloc(MESSAGE_CHUNK_SIZE);
    if( !data ) 
      return DISPATCH_SESSION_ERROR;

    assert(jsock_active == client || jsock_active == server);

    ret = js_read(jsock_active, data, MESSAGE_CHUNK_SIZE);
    if(ret < 0)
    {
      jfree(data);
      D_LOG(MSG_ERROR, "Read Error");
      return DISPATCH_SESSION_ERROR;
    }
    else if(ret == 0)
    {
      jfree(data);  
      D_LOG(MSG_DEBUG, "Peer Close The Connection");
      return DISPATCH_SESSION_CLOSE;
    }
    else // data read
    {
      /* now on ..we lose the ownership of data buffer */
      if( jsock_active == client ) // client 
      {
        ret = dispatch_client_data(data, (size_t)ret, session);
        if( !ret )
          return DISPATCH_SESSION_ERROR;
      }
      else // server
      {
        ret = dispatch_server_data(data, (size_t)ret, session);
        if( !ret )
          return DISPATCH_SESSION_ERROR;
      }     
    } // data read
  } // while()

  return 1;
}
#undef MESSAGE_CHUNK_SIZE


/*  we must inizialize the session (one per process) 
 *  if init_proxy_session failed, the caller must close the j_socket
 *  if success the session take ownership of the socket
 */
proxy_session_t *init_proxy_session(j_socket *jsock_client)
{
  int err;
  proxy_session_t *session = &g_session;

  struct sockaddr_in server_peer;
  j_socket *server;
  
  err = init_session(session);
  if( !err )
    return NULL;  


  err = try_to_resolv_host(&server_peer, jsock_client, session);
  if( !err )
  {
    fini_session(session);
    return NULL;
  }
  D_LOG(MSG_DEBUG, "Address of Remote Server: %s:%d", 
    inet_ntoa(server_peer.sin_addr), ntohs(server_peer.sin_port));
 
  
  server = js_init_new_sock_tcp();
  if( !server )
  {
    fini_session(session); 
    return NULL;
  } 

  err = js_connect(server, &server_peer, 0);
  if( err < 0 )
  {
    D_LOG(MSG_ERROR, "Faild to Connect To Remote Server: %s:%d", 
                     inet_ntoa(server_peer.sin_addr), ntohs(server_peer.sin_port));
    
    js_destroy_sock_tcp(server);
    fini_session(session);
    return NULL;
  }
    
  /* set socket and append data pointer/len */
  session->client_link = jsock_client;
  session->server_link = server;

  parse_client_message(session);
  
  return session;
}


int decoy_session(proxy_session_t *session)
{
  char *answer = "HTTP/1.1 404 Not found\r\nContent-Length: 9\r\n\r\nNOT FOUND";
  df_t* df = alloc_df(strdup(answer), strlen(answer), DF_HEADER, 0);
  if( !df )
    return 0;

  forward_to_client(session, df);
  return 1;
}

/* called if init_proxy_session() completes successfully */
int stream_proxy_session(proxy_session_t *session, unsigned int mode)
{
  int ret;
  j_socket_set_t *active;

  if( mode == J_MODE_DECOY )
  {
    decoy_session(session);
    fini_session(session);
    return 0;
  }
  else if( mode == J_MODE_REJECT )
  {
    D_LOG(MSG_DEBUG, "Unsupported Mode [Reject]: Closing Connection.. ");
    fini_session(session);
    return 0;
  }
  else if( mode == J_MODE_PROXY )
  {
    D_LOG(MSG_DEBUG, "Proxy Session Mode [Proxy]");
  }

  // inject mode...
  
  while(1)
  { 
    active = multiplex_session(session);
    if( IS_IO_TIMEOUT(active) )
    {
      D_LOG(MSG_DEBUG, "Timout on Session Multiplexing");
    }
    else if( active == NULL )
    {
      /* TODO: memleak in multiplex_session: if error we loos original static client/server set */
      D_LOG(MSG_ERROR, "Error on Session Multiplexing");
      fini_session(session);
      return 0;
    }
    else
    {
      ret = dispatch_session_event(active, session);
      if( ret == DISPATCH_SESSION_ERROR )
      {
        D_LOG(MSG_ERROR, "Error During Event Dispatching");
        release_j_socket_set(active);
        fini_session(session);
        return 0;
      }
      else if( ret == DISPATCH_SESSION_CLOSE )
      {
        D_LOG(MSG_DEBUG, "Session Close");
        release_j_socket_set(active);
        fini_session(session);
        return 1; 
      }
    } // event ok
      
  } // while(1)
  
  fini_session(session);
  return 1;
}
