#include <stdio.h>
#include <string.h>
#include "line_buffer.h"
#include "jmalloc.h"


/* @dst is raw memory, @search e classical 0 terminated string
 * @return: if search is found return ptr after it (NOTE: this can be out of boundary check it
 *          if search is not found return NULL
 */
static const char *memstr(const char* dst, const char *search, size_t size)
{
  size_t i;
  size_t search_len = strlen(search);

  if( size < search_len )
    return NULL;

  for(i=0; i <= size-search_len; i++)
  {
    if( !memcmp(dst+i, search, search_len) )
      return dst+i+search_len;  
  }

  return NULL;
}


line_buffer_t *create_lb(size_t size, const char* line_terminator)
{
  line_buffer_t *lb = jmalloc(sizeof(line_buffer_t)); 
  if( !lb )
    return NULL;

  memset(lb, 0x00, sizeof(line_buffer_t));

  lb->ln_term = line_terminator;
  lb->ln_buff = jmalloc(size);
  if( !lb->ln_buff )
  {
    jfree(lb);
    return NULL;
  }

  lb->ln_buff_size = size;
  lb->used = 0;

  return lb;
}

void destroy_lb(line_buffer_t *lb)
{
  assert( lb && lb->ln_buff );
  jfree(lb->ln_buff);
  jfree(lb);
}


char *lb_query_line(line_buffer_t *lb, size_t *len)
{
  const char *pos = memstr(lb->ln_buff, lb->ln_term, lb->used);
  if( !pos )
  {
    if( len )
      *len = 0;
    return NULL;
  }
  else /* there is at last  a valid line in the line buffer */
  {
    size_t new_len = pos - lb->ln_buff;
    char *out_buff = jmalloc(new_len);
    memset(out_buff, 0x00, new_len);

    memcpy(out_buff, lb->ln_buff, new_len);
    
    memmove(lb->ln_buff, lb->ln_buff + new_len, lb->used - new_len);
    lb->used = lb->used - new_len;

    /* for debug .. and gdb print facility */
    memset(lb->ln_buff + lb->used, 0x00, lb->ln_buff_size - lb->used);
    
    if( len )
      *len = new_len;

    return out_buff;
  }
}

/* return NULL if there is no bytes avaliable .. */
char *lb_query_last(line_buffer_t *lb, size_t *len)
{
  char *out_buff;  
  if( !lb->used )
  {
    if ( len )
      *len = 0;
    return NULL;
  }
  out_buff= jmalloc(lb->used);
  if( !out_buff )
    return NULL;

  memcpy(out_buff, lb->ln_buff, lb->used);
  
  /* debug */
  memset(lb->ln_buff, 0x00, lb->ln_buff_size);

  *len = lb->used;  
  lb->used = 0;
   
  return out_buff;
}

/* return data left .. and size in *discard */
const char *lb_insert_data(line_buffer_t *lb, const char* data, size_t len, size_t *discarded)
{
  size_t new_len = len;   
  
  if( new_len > lb->ln_buff_size - lb->used )
    new_len = lb->ln_buff_size - lb->used;

  memcpy(lb->ln_buff + lb->used, data, new_len);
  lb->used += new_len;

  if( discarded )
    *discarded = len - new_len; 

  return data+new_len;
}
